/* [KSC4]-------------------------------------------------------------------[]
 *
 *  MAIN.C - Main module for the Power Glitch Monitor, version 0.4
 *           Written by Marco "Kiko" Carnut, 2008-Jan-04
 *
 *  Portability: Tested with avr-gcc 4.2.1 and avr-libc 1.4.6 with an ATmega8
 *  ------------ device.
 *
 * []-----------------------------------------------------------------------[]
 */

#ifndef __AVR_ATmega8__
#error Hey, this code was only tested in an ATmega8
#endif

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include <string.h>
#include <stdlib.h>

#define F_CPU                8000000L

#include <util/delay.h>

//#define HAVE_XTRA_CMDS
#define HAVE_LCD
#define HAVE_TWI

#ifdef HAVE_TWI
#include <util/twi.h>
#endif // HAVE_TWI

/* --------------------------- Macro-Defined Constants -------------------- */

#define ADC_CHANNELS               6
#define MAXTXFIFOLENGTH          120

#define TRUE                       1
#define FALSE                      0

#define FLAG_SERIAL_ECHO_OVERRIDE  1
#define FLAG_HAS_SECOND            4
#define FLAG_LCD_ECHO              8
#define FLAG_HAS_ANOMALY          16
#define FLAG_SERIAL_ECHO          32
#define FLAG_SERIAL_LOG_ENABLED   64
#define FLAG_CLOCK_ADJUST        128

#define BUZZER_DDR                 DDRB
#define BUZZER_PORT                PORTB
#define BUZZER_PIN                 PB1

#define MAXARGS                    3

#define TWI_PAGE_SIZE            128

/* ---------------------------- Function Prototypes ----------------------- */

void xmit_str(char *str);
void xmit_hex(unsigned short val, unsigned char ndigits);
void xmit_int_dp(unsigned short val, unsigned char ndigits, char point);
void dec2digits(unsigned short val, unsigned char ndigits, char point);
void xmit_clock(volatile unsigned char *c, unsigned char max, unsigned char month_name);
void xmit_str_p(prog_char *str);
void xmit(char c);
void do_xmit(char c);
void handle_serial_input(void);
void update_elapsed(unsigned char *c);
void xmit_elapsed(unsigned char *c);
void beep(unsigned short cycles);
void do_xmit_str_p(prog_char *str);
void update_clock(volatile unsigned char *c, unsigned char inc0);

void twi_retry(void);
void twi_fail(void);
void twi_eeprom_write_byte(unsigned short address, unsigned char data);
void twi_eeprom_read_block(unsigned short address, unsigned char *buf, unsigned char len);
void twi_eeprom_write_page(unsigned short address, unsigned short start);
void twi_eeprom_read_byte(unsigned short address);
unsigned char twi_wait_until_idle(unsigned char ms);

void xmit_hex_nibble(unsigned char nibble);
void xmit_dump(void);
unsigned char twi_is_busy(void);

void start_new_event(unsigned char type);
void finish_event(void);

void lcd_init(unsigned char dispAttr);
void lcd_puts_p(const prog_char *p);
void lcd_home(void);
void lcd_putc(char c);
void lcd_write(unsigned char data, unsigned char rs);

void execute_command(void);
void cmd_date(unsigned char argc, char **argv);
void cmd_time(unsigned char argc, char **argv);
void cmd_dump(unsigned char argc, char **argv);
void set_time_date(unsigned char argc, char **argv, unsigned char base);
void cmd_log(unsigned char argc, char **argv);
void cmd_clear(unsigned char argc, char **argv);
void cmd_start(unsigned char argc, char **argv);
void cmd_finish(unsigned char argc, char **argv);
void cmd_event(unsigned char argc, char **argv);
void cmd_poke(unsigned char argc, char **argv);
void cmd_fill(unsigned char argc, char **argv);
void cmd_reboot(unsigned char argc, char **argv);
void cmd_param(unsigned char argc, char **argv);
void cmd_ver(unsigned char argc, char **argv);

struct event_st *get_event(short e);
void show_usage(prog_char *p);

/* ---------------------- Global Variables and Declarations --------------- */

volatile char txfifo[MAXTXFIFOLENGTH];
volatile unsigned char fifo_head;
volatile unsigned char fifo_tail;

typedef union {
    volatile unsigned short as_u16[2];
    volatile unsigned long  as_u32;
} u32_16; 

volatile u32_16 total_ticks;

volatile unsigned long t1_ticks;
volatile unsigned char semicycles;
volatile unsigned long slow_semicycles,fast_semicycles;
volatile unsigned long sag_semicycles,surge_semicycles;

short params[] = {
    0,          // PARAM_FLAGS
    450,        // PARAM_ADC_MIN
    766,        // PARAM_ADC_SAG
    847,        // PARAM_ADC_SURGE
    937,        // PARAM_ICR_MIN
    1146,       // PARAM_ICR_MAX
    1000,       // PARAM_BEEP_POWERFAIL
    250,        // PARAM_BEEP_POWEROK
    2,          // PARAM_BEEP_SAG
    2,          // PARAM_BEEP_SURGE
    2,          // PARAM_BEEP_LOW_HZ
    2           // PARAM_BEEP_HIGH_HZ
};

#define PARAM_FLAGS                          0
#define PARAM_ADC_MIN                        1
#define PARAM_ADC_SAG                        2
#define PARAM_ADC_SURGE                      3
#define PARAM_ICR_MIN                        4
#define PARAM_ICR_MAX                        5
#define PARAM_BEEP_POWERFAIL                 6
#define PARAM_BEEP_POWEROK                   7
#define PARAM_BEEP_SAG                       8
#define PARAM_BEEP_SURGE                     9
#define PARAM_BEEP_LOW_HZ                   10
#define PARAM_BEEP_HIGH_HZ                  11

#define PARAM_FLAGS_RECORD_FREQ_ANOMALIES    1
#define PARAM_FLAGS_RECORD_VOLT_ANOMALIES    2

//short adc_min=450;
short adc_sag=766,adc_surge=847;
short icr_min=937,icr_max=1146;

volatile unsigned char flags;
volatile unsigned char has_lcd_update;

volatile unsigned short voltage;
volatile unsigned char adc_count;

enum anomaly_t {
    ANOMALY_NONE           = 0,
    ANOMALY_POWER_OUTAGE   = 1,
    ANOMALY_VOLTAGE_SAG    = 2,
    ANOMALY_VOLTAGE_SURGE  = 3,
    ANOMALY_FREQUENCY_HIGH = 4,
    ANOMALY_FREQUENCY_LOW  = 8
} anomaly_type,prev_anomaly_type;

// The maximum values for each of those counters
unsigned char limits[] PROGMEM = { 99,59,59,23,31,12,99 };

// The starting values for each counter
unsigned char base[] PROGMEM = { 0,0,0,0,1,1,0 };

#define SUBSEC 0
#define SEC    1
#define MIN    2
#define HOUR   3
#define DAY    4
#define MONTH  5
#define YEAR   6

volatile unsigned char rtc_crystal[11];
volatile unsigned char rtc_mains[11];
volatile unsigned char dozens;
volatile unsigned char millissecs;
unsigned char rtc_last_failure[11];
unsigned char rtc_elapsed[4];
unsigned char cmd_set_clock_index,cmd_temp;

char digits[7];

unsigned short nominal_voltage=220;

#ifdef HAVE_TWI

// TWI EEPROM

volatile enum {
    TWI_IDLE,            // a
    TWI_SENDING_START,   // b
    TWI_SENDING_RESTART, // c
    TWI_SENDING_SLA_R,   // d
    TWI_SENDING_SLA_W,   // e
    TWI_SENDING_LO_ADDR, // f
    TWI_SENDING_HI_ADDR, // g
    TWI_SENDING_DATA,    // h
    TWI_READING_DATA,    // i
    TWI_SUCCESS,         // j
    TWI_ABORT,           // k
    TWI_FAIL             // l
} twi_state,twi_fail_state;

enum {
    TWI_WRITE, TWI_READ, TWI_PAGE_WRITE
} twi_op;

unsigned short twi_eeprom_addr;
unsigned char  twi_eeprom_data;
unsigned char  twi_retry_count;
unsigned char  twi_guard;

unsigned short twi_eeprom_data_index;
unsigned char  twi_eeprom_data_len;
unsigned char *twi_eeprom_data_ptr;

#ifdef HAVE_XTRA_CMDS

unsigned short hexdump_addr,hexdump_end;
enum {
    HEXDUMP_IDLE,
    HEXDUMP_XMIT_SPACE,
    HEXDUMP_XMIT_HI_DATA_NIBBLE,
    HEXDUMP_XMIT_LO_DATA_NIBBLE,
    HEXDUMP_XMIT_CR,
    HEXDUMP_XMIT_LF
} hexdump_state;

#endif // HAVE_XTRA_CMDSs

#endif // HAVE_TWI

#ifdef HAVE_LCD

/* display on/off, cursor on/off, blinking char at cursor position */
#define LCD_DISP_OFF             0x08   /* display off                            */
#define LCD_DISP_ON              0x0C   /* display on, cursor off                 */
#define LCD_DISP_ON_BLINK        0x0D   /* display on, cursor off, blink char     */
#define LCD_DISP_ON_CURSOR       0x0E   /* display on, cursor on                  */
#define LCD_DISP_ON_CURSOR_BLINK 0x0F   /* display on, cursor on, blink char      */

#define LCD_LINES           4     /* visible lines */
#define LCD_COLUMNS        20     /* visible columns */

char lcd[LCD_LINES*LCD_COLUMNS];
volatile unsigned char lcd_current;
unsigned char lcd_cursor;

#define enable_lcd_echo()            (flags |=  FLAG_LCD_ECHO)
#define disable_lcd_echo()           (flags &= ~FLAG_LCD_ECHO)
#define is_lcd_echo_enabled()        (flags &   FLAG_LCD_ECHO)

#define enable_serial_echo()         (flags |=  FLAG_SERIAL_ECHO)
#define disable_serial_echo()        (flags &= ~FLAG_SERIAL_ECHO)
#define is_serial_echo_enabled()     (flags &   FLAG_SERIAL_ECHO)

#define enable_serial_override()     (flags |=  FLAG_SERIAL_ECHO_OVERRIDE)
#define disable_serial_override()    (flags &= ~FLAG_SERIAL_ECHO_OVERRIDE)
#define is_serial_override_enabled() (flags &   FLAG_SERIAL_ECHO_OVERRIDE)

#endif // HAVE_LCD

char month_names[] PROGMEM = "Jan\0Feb\0Mar\0Apr\0May\0Jun\0"
                             "Jul\0Aug\0Sep\0Oct\0Nov\0Dec";

unsigned char ndays_in_month[] PROGMEM = { 
    31,28,31,30,31,30,31,31,30,31,30,31
};

unsigned short beep_cycles;

#define MAXCMDLINELENGTH     30

char cmdline[MAXCMDLINELENGTH];
unsigned char cmdline_cursor,cmd_in_execution;

#ifdef HAVE_XTRA_CMDS

char cmdnames[] PROGMEM = "date\0time\0reboot\0log\0clear\0event\0param\0ver\0"
    "start\0finish\0poke\0fill\0dump\0";

void (*cmdvects[])(unsigned char, char **) PROGMEM = {
    cmd_date,cmd_time,cmd_reboot,cmd_log,cmd_clear,cmd_event,cmd_param,cmd_ver,
    cmd_start,cmd_finish,cmd_poke,cmd_fill,cmd_dump
};

#else

char cmdnames[] PROGMEM = "date\0time\0reboot\0log\0clear\0event\0param\0ver\0";

void (*cmdvects[])(unsigned char, char **) PROGMEM = {
    cmd_date,cmd_time,cmd_reboot,cmd_log,cmd_clear,cmd_event,cmd_param,cmd_ver
};

#endif // HAVE_XTRA_CMDS

// This defines an event to be recorded. Its size must be a multiple of two
// The way it is defined here, it's highly suboptimal -- several fields could
// be considerably shorter.

struct event_st {
    unsigned char  type;
    unsigned char  timestamp[7];
    unsigned char  duration[4];
    unsigned short voltage;
    unsigned short frequency;
    unsigned long  fast_semicycles;
    unsigned long  slow_semicycles;
    unsigned long  sag_semicycles;
    unsigned long  surge_semicycles;
};

#define EVENT_QUEUE_OVERFACTOR      2
#define MAX_UNFLUSHED_EVENTS        (TWI_PAGE_SIZE/sizeof(struct event_st))
#define EVENT_QUEUE_SIZE            (EVENT_QUEUE_OVERFACTOR*MAX_UNFLUSHED_EVENTS)
#define TWI_BUF_SIZE                (EVENT_QUEUE_OVERFACTOR*TWI_PAGE_SIZE)

union {
    struct event_st event[EVENT_QUEUE_SIZE];
    unsigned char twi_page_buf[TWI_BUF_SIZE];
} event_queue;

unsigned char event_tail;
unsigned char unflushed_events;
unsigned char first_unflushed;
unsigned char within_event;
short nevents;

unsigned char grace_time;

/* -------------------------- Macro-Defined Functions --------------------- */

#define xmit_pstr(str)       xmit_str_p(PSTR(str))
#define do_xmit_pstr(str)    do_xmit_str_p(PSTR(str))
#define usage(str)           show_usage(PSTR(str))
#define lcd_pstr(s)          lcd_puts_p(PSTR(s))

#define tx_busy()            (!(UCSRA  & _BV(UDRE)))

#define adc_busy()           (!(ADCSRA & _BV(ADIF)))
#define adc_start()          ADCSRA |= _BV(ADSC)
#define set_adc_channel(c)   \
    (ADMUX = ((ADMUX & ~(_BV(MUX3)|_BV(MUX2)|_BV(MUX1)|_BV(MUX0))) \
    |((c)<<MUX0)))
#define adc_value()          (ADCL|((ADCH & 3)<<8))

#define xmit_int(a,b)        xmit_int_dp((a),(b),-1)

#define xmit_interrupt_enable() \
    (UCSRB |= _BV(UDRIE))
#define xmit_interrupt_disable() \
    (UCSRB &= ~_BV(UDRIE))

#define analog_comparator_bit() ((ACSR & _BV(ACO))==_BV(ACO))

/* -------------------------------- Main Program -------------------------- */

int main(void)
{
    unsigned char n;
//  DDRC |= _BV(PC5);
    BUZZER_DDR |= _BV(BUZZER_PIN);

    /* UART initialization */
    UCSRB  =  _BV(TXEN) | _BV(RXEN) // Enable transmitter and receiver,
           |  _BV(RXCIE);           // and receiver interrupts
    UBRRL  = 51;                    // Set 9600bps for internal 8MHz RC osc

    /* Timer1 Setup -- handles waveform capture: freq & volt measurement */
    TCCR1B  = _BV(ICNC1)            // ICP noise canceler on
            | 3;                    // prescaler = 64, clk(T1) = 125,000 Hz

    OCR1A = 600;

    /* Timer2 Setup -- handles the crystal-based real time clock */
    TCCR2 = _BV(WGM21)              // Enable CTC mode
          | 4;                      // perscaler = 64, clk(T2) = 125,000Hz
    OCR2 = 124;                     // Divide by 125, OC match each 1000Hz

    TIMSK  |= _BV(TICIE1)           // Enable Input Capture interrupt
           |  _BV(OCIE1A)           // Enable Output Compare 1A interrupt
           |  _BV(OCIE2);           // Enable Timer2 Output Compare (RTC)

    /* Disable Analog Comparator, we don't use it for anything */
    ACSR |= _BV(ACD);

    /* TWI initialization */
    PORTC |= _BV(PC4) | _BV(PC5);   // Enable Pull-ups on SCL and SDA
    TWSR = 0;
    TWBR = 10;

#ifdef HAVE_LCD
    for (n=0;n<sizeof(lcd);n++) lcd[n]=' ';
    lcd_init(LCD_DISP_ON);
#endif

    // Initialize RTC
    for (n=0;n<7;n++) {
        rtc_crystal[n]=rtc_mains[n]=pgm_read_byte(base+n);
    }
    nevents=-1;
    grace_time=240;

    enable_serial_echo();
    xmit_str("\r\n\r\n");
    cmd_ver(0,NULL);
    xmit_str("\r\n> ");
    for (;;) {
        /* Enable interrupts */
        sei();
        set_sleep_mode(SLEEP_MODE_IDLE);
        sleep_mode();
        asm("nop");
        if (flags & FLAG_HAS_SECOND) {
            flags &= ~FLAG_HAS_SECOND;
            if (flags & FLAG_SERIAL_LOG_ENABLED)
                enable_serial_echo();
            else
                disable_serial_echo();
            long c1=*((unsigned long *)(rtc_crystal+7));
            long c2=*((unsigned long *)(rtc_mains+7));
            short drift=c1-c2;
            if (drift<0) {
                xmit('-'); drift=-drift;
            } else {
                xmit('+');
            }
            xmit_int_dp(drift,5,1);
            xmit(' ');
            if ((anomaly_type & 3) == ANOMALY_POWER_OUTAGE) {
                enable_lcd_echo();
                lcd_cursor=0;
                xmit_pstr("* AC POWER OFFLINE *");
                disable_lcd_echo();
                xmit(' ');
            } else {
                unsigned long base=62500-520*(120-semicycles);
                base*=60000;
                unsigned short f=base/(t1_ticks>>1);
                xmit_int_dp(semicycles,4,-1); xmit(' ');
                semicycles=0;
                t1_ticks=0;

                lcd_cursor=0; enable_lcd_echo();
                xmit_int_dp(nominal_voltage,3,-1);
                xmit(':'); xmit(' ');
                u32_16 v;
#ifndef CALLIBRATION
                v.as_u32=(unsigned long)voltage*2978;
                xmit_int_dp(v.as_u16[1],4,1);
                xmit_pstr("V ");
#else
                xmit_int_dp(voltage,5,1);
                xmit_pstr("U  ");
                xmit_int_dp(adc_count,3,-1);
                xmit_str("  OCR1A=");
                xmit_int_dp(OCR1A,4,-1);
                xmit_str("  ");
#endif // CALLIBRATION
                xmit_int_dp(f,5,3);
                xmit_pstr("Hz ");
                disable_lcd_echo();
                xmit_int_dp(fast_semicycles,5,-1);
                xmit(' ');
                xmit_int_dp(slow_semicycles,5,-1);
                xmit(' ');
                xmit_int_dp(surge_semicycles,5,-1);
                xmit(' ');
                xmit_int_dp(sag_semicycles,5,-1);
                xmit(' ');
                voltage=0;
                adc_count=0;
            }
            lcd_cursor=41;
            enable_lcd_echo();
            xmit_clock(rtc_crystal,6,1);
            if (nevents>=0) {
                xmit(' ');
                xmit('0'+anomaly_type);
                xmit(' ');
                xmit_int_dp(nevents+1,4,-1);
                xmit_pstr(": ");
                xmit_elapsed(rtc_elapsed);
                xmit_clock(rtc_last_failure,7,0);
            }
            disable_lcd_echo();
            xmit_pstr("\r\n");
        }
        if (anomaly_type != ANOMALY_NONE) {
            if (prev_anomaly_type) {
                if (prev_anomaly_type != anomaly_type) {
                    finish_event();
                    start_new_event(anomaly_type);
                    prev_anomaly_type=anomaly_type;
                }
            } else {
               start_new_event(anomaly_type);
               prev_anomaly_type=anomaly_type;
            }
        } else {
            if (prev_anomaly_type) {
                finish_event();
                prev_anomaly_type=ANOMALY_NONE;
            }
        }
        if (cmd_in_execution==1) execute_command();
#if defined(HAVE_TWI) && defined(HAVE_XTRA_CMDS)
        if (hexdump_state!=HEXDUMP_IDLE) xmit_dump();
#endif
#ifdef HAVE_LCD
        if (has_lcd_update) {
            has_lcd_update=0;
            if (lcd_current==sizeof(lcd)) {
                lcd_current=0; lcd_home();
            } else {
                lcd_write((unsigned char)lcd[(int)lcd_current++],1);
            }
        }
#endif // HAVE_LCD
    }

    /* We never return, but we put it here to avoid gcc warnings */
    return 0;
}

void start_new_event(unsigned char type)
{
    if (within_event || grace_time) return;
    within_event=TRUE; nevents++;
    struct event_st *e=event_queue.event+event_tail;
    e->type=type;
    memset(rtc_elapsed,0,sizeof(rtc_elapsed));
    memcpy(e->timestamp,(unsigned char *)rtc_crystal,sizeof(rtc_crystal));
    memcpy(rtc_last_failure,(unsigned char *)rtc_crystal,sizeof(rtc_crystal));
    if ((type & 3)==ANOMALY_POWER_OUTAGE) {
        beep(params[PARAM_BEEP_POWERFAIL]);
    }
    //grace_time=60;
}

void finish_event(void)
{
    if (!within_event) return;
    struct event_st *e=event_queue.event+event_tail;
    e->fast_semicycles=fast_semicycles;
    e->slow_semicycles=slow_semicycles;
    e->sag_semicycles=sag_semicycles;
    e->surge_semicycles=surge_semicycles;
    memcpy(e->duration,rtc_elapsed,sizeof(rtc_elapsed));
    if (++event_tail==EVENT_QUEUE_SIZE) event_tail=0;
#ifdef HAVE_TWI
    if (++unflushed_events==MAX_UNFLUSHED_EVENTS) {
        unsigned short addr,start;
        addr  = (nevents+1-MAX_UNFLUSHED_EVENTS)*sizeof(struct event_st);
        start = first_unflushed*sizeof(struct event_st);
        if (twi_wait_until_idle(3)) {  // It might be reading right now
            twi_eeprom_write_page(addr,start);
            unflushed_events=0;
            first_unflushed+=MAX_UNFLUSHED_EVENTS;
            if (first_unflushed>=EVENT_QUEUE_SIZE) first_unflushed-=EVENT_QUEUE_SIZE;
        }
    }
#endif // HAVE_TWI
    if ((e->type & 3)==ANOMALY_POWER_OUTAGE) beep(params[PARAM_BEEP_POWEROK]);
    within_event=FALSE;
}

struct event_st *get_event(short e)
{
    short first,last;
    last=nevents;
    if (e<0 || e>last) return NULL;
    first=last+1-EVENT_QUEUE_SIZE;
    if (first<0) first=0;
#ifdef HAVE_TWI
    if (e<first) { // Too old to be in the queue, get it from the TWI EEPROM
        unsigned short addr=e*sizeof(struct event_st);
        //do_xmit_pstr("From EEPROM\r\n");
        static struct event_st evt;
        if (!twi_wait_until_idle(12))  // It might be writing right now
            return NULL;
        twi_eeprom_read_block(addr,(unsigned char *)&evt,sizeof(evt));
        return &evt;
    }
#endif // HAVE_TWI
    if (e>=first) { // It's still in the queue, just return it
        e=e-last-1+event_tail;
        if (e<0) e+=EVENT_QUEUE_SIZE;
        //do_xmit_pstr("From cache item: "); do_xmit('0'+e); do_xmit_pstr("\r\n");
        return event_queue.event+e;
    }
    return NULL;
}

void update_clock(volatile unsigned char *c, unsigned char inc0)
{
    if (flags & FLAG_CLOCK_ADJUST) return;
    unsigned char i;
    unsigned char inc=inc0;
    for (i=0;i<7;i++) {
        unsigned char limit=pgm_read_byte(limits+i);
        if (i==DAY) {
            // Handle varying month lengths and leap years
            // This will work up to 2099
            limit  = pgm_read_byte(c[MONTH]+ndays_in_month-1);
            limit += (c[MONTH]==2) && !(c[YEAR]&3);
        }
        c[i]+=inc;
        if (c[i]<=limit) break;
        c[i]=pgm_read_byte(base+i);
        inc=1;
    }
    // Increment linear counter as well -- simplifies
    // calculating time differences
    unsigned long *p=(unsigned long *)(c+7);
    (*p)+=inc0;
}

void update_elapsed(unsigned char *c)
{
    unsigned char i;
    for (i=0;i<4;i++) {
        unsigned char limit=pgm_read_byte(limits+i);
        if (i==HOUR) limit=255;
        if (++c[i]<=limit) break;
        c[i]=0;
    }
}

void xmit_clock(volatile unsigned char *c, unsigned char max, unsigned char month_name)
{
    prog_char *sep =  PSTR(".:: --");
    unsigned char n;
    for (n=0;n<max;n++) {
        unsigned char s=pgm_read_byte(sep+6-n);
        if (s) xmit(s);
        if (n==1 && month_name) {
            xmit_str_p(month_names+((c[5]-1)<<2));
        } else {
            xmit_int(c[6-n],2);
        }
    }
}

void xmit_elapsed(unsigned char *c)
{
    xmit_int_dp(c[HOUR],3,-1);
    xmit('h');
    xmit_int_dp(c[MIN],2,-1);
    xmit('m');
    xmit_int_dp(c[SEC],2,-1);
    xmit('s');
    xmit_int_dp(c[SUBSEC],2,-1);
}

SIGNAL(SIG_OUTPUT_COMPARE1A)
{
    ADMUX  = 0;                     // External reference, channel 0
    ADCSRA = _BV(ADEN) | _BV(ADIE)  // Enable ADC, enable ADC interrupt
           | _BV(ADSC)              // Start ADC
           | 6;                     // prescaler = 64, clk(ADC) = 125KHz
}

SIGNAL(SIG_INPUT_CAPTURE1)
{
    TCNT1=0;
    if (grace_time) grace_time--;
    t1_ticks+=ICR1;
    semicycles++;
    if (++dozens==12) {
        update_clock(rtc_mains,10);
        dozens=0;
    }
    enum anomaly_t anomaly=ANOMALY_NONE;
    if (ICR1>params[PARAM_ICR_MAX]) {
        slow_semicycles++; beep(params[PARAM_BEEP_LOW_HZ]);
        if (params[PARAM_FLAGS] & PARAM_FLAGS_RECORD_FREQ_ANOMALIES) 
            anomaly=ANOMALY_FREQUENCY_LOW;
    } else if (ICR1<params[PARAM_ICR_MIN]) {
        fast_semicycles++; beep(params[PARAM_BEEP_HIGH_HZ]);
        if (params[PARAM_FLAGS] & PARAM_FLAGS_RECORD_FREQ_ANOMALIES) 
            anomaly=ANOMALY_FREQUENCY_HIGH;
    }
    if (!grace_time)
        anomaly_type = (anomaly_type & 0xF3) | anomaly;
}

SIGNAL(SIG_ADC)
{
    unsigned short adc=ADC;
    voltage+=adc>>1;
    ADCSRA = 0; // Disable ADC entirely
    adc_count++;
    enum anomaly_t anomaly=ANOMALY_NONE;
    if (adc<params[PARAM_ADC_MIN]) {
        anomaly=ANOMALY_POWER_OUTAGE;
        TIMSK &= ~_BV(TICIE1);
    } else {
        if (adc<params[PARAM_ADC_SAG]) {
            sag_semicycles++; beep(params[PARAM_BEEP_SAG]);
            if (params[PARAM_FLAGS] & PARAM_FLAGS_RECORD_FREQ_ANOMALIES) 
                anomaly=ANOMALY_VOLTAGE_SAG;
        } else if (adc>params[PARAM_ADC_SURGE]) {
            surge_semicycles++; beep(params[PARAM_BEEP_SURGE]);
            if (params[PARAM_FLAGS] & PARAM_FLAGS_RECORD_FREQ_ANOMALIES) 
                anomaly=ANOMALY_VOLTAGE_SURGE;
        } else {
        }
    }
    if (!grace_time)
        anomaly_type = (anomaly_type & 0xFC) | anomaly;
    if (!(TIMSK & _BV(TICIE1))) {
        TCNT1=0; semicycles=0;
        if (grace_time) grace_time--;
        if (anomaly==ANOMALY_NONE) { 
            TIMSK |= _BV(TICIE1); grace_time=1;
        }
    }
#ifdef CALLIBRATION
    PORTC &= ~_BV(PC0);
    DDRC  |=  _BV(PC0);
    _delay_us(10);
    DDRC  &= ~_BV(PC0);
#endif // CALLIBRATION
}

SIGNAL(SIG_OUTPUT_COMPARE2)
{
    if (++millissecs==10) {
        millissecs=0;
        update_clock(rtc_crystal,1);
        if (!rtc_crystal[SUBSEC]) flags |= FLAG_HAS_SECOND;
        if (within_event) update_elapsed(rtc_elapsed);
    }
    has_lcd_update=1;
    if (beep_cycles) {
        if (!--beep_cycles) BUZZER_PORT &= ~_BV(BUZZER_PIN);
    }
#ifdef HAVE_TWI
    if (twi_guard) {
        if (!--twi_guard) {
            twi_state=TWI_IDLE;
            TWCR &= ~_BV(TWEN);
        }
    }
#endif // HAVE_TWI
}

void beep(unsigned short cycles)
{
    if (!cycles) return;
    if (cycles<beep_cycles) return;
    BUZZER_PORT |= _BV(BUZZER_PIN);
    beep_cycles=cycles;
}

void dec2digits(unsigned short val, unsigned char ndigits, char point)
{
    unsigned char n;
    if (point>=0) {
        ndigits++;
        point=ndigits-point;
    }
    for (n=ndigits;n>0;n--) {
        if (n==point) {
            digits[n-1]='.';
        } else {
            digits[n-1]=val%10+'0';
            val/=10;
        }
    }
    digits[ndigits]=0;
}

void xmit_int_dp(unsigned short val, unsigned char ndigits, char point)
{
    dec2digits(val,ndigits,point);
    xmit_str(digits);
}

void xmit_hex(unsigned short val, unsigned char ndigits)
{
    char d[5];
    unsigned char n;
    for (n=ndigits;n>0;n--) {
        char c=(val&0xF)+'0';
        if (c>'9') c+=7;
        d[n-1]=c;
        val>>=4;
    }
    d[ndigits]=0;
    xmit_str(d);
}

void xmit_hex_str(unsigned char *p, int length)
{
    while (length--) xmit_hex(*p++,2);
}

SIGNAL(SIG_UART_DATA)
{
    if (fifo_tail==fifo_head) {
        xmit_interrupt_disable();
    } else {
        UDR = txfifo[fifo_tail++];
        if (fifo_tail==MAXTXFIFOLENGTH) fifo_tail=0;
    }
}

SIGNAL(SIG_UART_RECV)
{
    unsigned char c=UDR;
    if (flags & FLAG_SERIAL_LOG_ENABLED) {
        flags &= ~FLAG_SERIAL_LOG_ENABLED;
        do_xmit_pstr("\r\n> ");
        return;
    }
#if defined(HAVE_TWI) && defined(HAVE_XTRA_CMDS)
    if (hexdump_state!=HEXDUMP_IDLE) {
        hexdump_state=HEXDUMP_IDLE;
        do_xmit_pstr("\r\n");
        return;
    }
#endif
    if (!cmd_in_execution) {
        if (c==13) {            // Enter
            cmd_in_execution=1; // Dispatcher
            do_xmit_pstr("\r\n");
        } else if (c==3) {
            do_xmit_pstr("\r\n> ");
            cmdline_cursor=0;
        } else if (c==8) {
            if (cmdline_cursor) {
                do_xmit_pstr("\b \b");
                cmdline_cursor--;
            }
        } else if (c>=' ' && c<=127) {
            if (cmdline_cursor<MAXCMDLINELENGTH) {
                cmdline[cmdline_cursor++]=c;
                do_xmit(c);
            }
        }
    }
}

void xmit_str(char *str)
{
    for (;*str;str++) xmit(*str);
}

void xmit_str_p(prog_char *str)
{
    for (;;) {
        unsigned c=pgm_read_byte(str++);
        if (!c) break;
        xmit(c);
    }
}

void do_xmit_str_p(prog_char *str)
{
    enable_serial_override();
    xmit_str_p(str);
    disable_serial_override();
}

void do_xmit(char c)
{
    unsigned char next_head=fifo_head+1;
    if (next_head==MAXTXFIFOLENGTH) next_head=0;
    /* If there is still room in the FIFO, add character by character */
    if (next_head==fifo_tail) return;
    txfifo[fifo_head]=c;
    fifo_head=next_head;
    xmit_interrupt_enable();
}

void xmit(char c)
{
    if (is_serial_echo_enabled() || is_serial_override_enabled())
        do_xmit(c);
#ifdef HAVE_LCD
    if (is_lcd_echo_enabled()) {
        lcd[lcd_cursor++]=c;
        if (lcd_cursor==LCD_COLUMNS) lcd_cursor=2*LCD_COLUMNS;
        else if (lcd_cursor==3*LCD_COLUMNS) lcd_cursor=LCD_COLUMNS;
        else if (lcd_cursor==2*LCD_COLUMNS) lcd_cursor=3*LCD_COLUMNS;
        else if (lcd_cursor==4*LCD_COLUMNS) lcd_cursor=0;
    }
#endif // HAVE_LCD
}

/* --------------- I2C (a.k.a TWI) Serial EEPROM Support ------------------ */

#ifdef HAVE_TWI

void twi_eeprom_write_page(unsigned short address, unsigned short start)
{
    if (twi_is_busy()) return;
    twi_eeprom_addr=address & ~(TWI_PAGE_SIZE-1);
    twi_eeprom_data_index=start;
    twi_eeprom_data_len=TWI_PAGE_SIZE;
    twi_state=TWI_SENDING_START;
    twi_retry_count=100;
    twi_op=TWI_PAGE_WRITE;
    twi_guard=0;
    // Send start condition
    TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN) | _BV(TWIE);
}

void twi_eeprom_inc_index(void)
{
    if (++twi_eeprom_data_index==TWI_BUF_SIZE)
       twi_eeprom_data_index=0;
    twi_eeprom_data_len--;
}

void twi_eeprom_read_block(unsigned short address, unsigned char *buf, unsigned char len)
{
    if (twi_is_busy()) return;
    twi_eeprom_addr=address;
    twi_state=TWI_SENDING_START;
    twi_retry_count=100;
    twi_op=TWI_READ;
    twi_guard=0;
    twi_eeprom_data_ptr=buf;
    twi_eeprom_data_len=len;
    // Send start condition
    TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN) | _BV(TWIE);
}

#ifdef HAVE_XTRA_CMDS

void twi_eeprom_read_byte(unsigned short address)
{
    if (twi_is_busy()) return;
    twi_eeprom_addr=address;
    twi_state=TWI_SENDING_START;
    twi_retry_count=100;
    twi_op=TWI_READ;
    twi_guard=0;
    twi_eeprom_data_ptr=&twi_eeprom_data;
    twi_eeprom_data_len=1;
    // Send start condition
    TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN) | _BV(TWIE);
}

void twi_eeprom_write_byte(unsigned short address, unsigned char data)
{
    if (twi_is_busy()) return;
    twi_eeprom_addr=address;
    twi_eeprom_data=data;
    twi_state=TWI_SENDING_START;
    twi_retry_count=100;
    twi_op=TWI_WRITE;
    twi_guard=0;
    // Send start condition
    TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN) | _BV(TWIE);
}

#endif // HAVE_XTRA_CMDS

unsigned char twi_wait_until_idle(unsigned char ms)
{
    unsigned char n;
    for (n=0;n<ms;n++) {
        cli();
        if (!twi_is_busy() && !twi_guard) {
            twi_guard=2; sei();
            return 1;
        }
        sei();
        _delay_ms(1);
    }
    do_xmit_pstr("Panic: TWI busy\n");
    return 0;
}

SIGNAL(SIG_2WIRE_SERIAL)
{
    //BUZZER_PORT ^= _BV(BUZZER_PIN);
    //do_xmit('a'+twi_state);
    if (twi_state==TWI_SENDING_START) {
        if (TW_STATUS == TW_MT_ARB_LOST) {
            twi_retry();
        } else if ( (TW_STATUS != TW_REP_START) && (TW_STATUS != TW_START)) {
            twi_fail();
        } else {
            twi_state=TWI_SENDING_SLA_W;
            TWDR = 0xA0;
            TWCR = _BV(TWINT) | _BV(TWEN) | _BV(TWIE);
        }
    } else if (twi_state==TWI_SENDING_SLA_W) {
        if (TW_STATUS == TW_MT_SLA_NACK) {
            twi_retry();
        } else if (TW_STATUS == TW_MT_ARB_LOST) {
            twi_retry();
        } else if (TW_STATUS != TW_MT_SLA_ACK) {
            // Error, send stop condition
            twi_fail();
        } else {
            // So far so good, let's send the data
            twi_state = TWI_SENDING_HI_ADDR;
            TWDR = twi_eeprom_addr>>8;
            TWCR = _BV(TWINT) | _BV(TWEN) | _BV(TWIE);
        }
    } else if (twi_state==TWI_SENDING_HI_ADDR) {
        if (TW_STATUS == TW_MT_DATA_NACK) {
            // Busy, send stop condition and abort
            twi_fail();
        } else if (TW_STATUS == TW_MT_ARB_LOST) {
            twi_retry();
        } else if (TW_STATUS != TW_MT_DATA_ACK) {
            // Error, send stop condition
            twi_fail();
        } else {
            // So far so good, let's send the data
            twi_state = TWI_SENDING_LO_ADDR;
            TWDR = twi_eeprom_addr;
            TWCR = _BV(TWINT) | _BV(TWEN) | _BV(TWIE);
        }
    } else if (twi_state==TWI_SENDING_LO_ADDR) {
        if (TW_STATUS == TW_MT_DATA_NACK) {
            // Busy, send stop condition and abort
            twi_fail();
        } else if (TW_STATUS == TW_MT_ARB_LOST) {
            twi_retry();
        } else if (TW_STATUS != TW_MT_DATA_ACK) {
            // Error, send stop condition
            twi_fail();
        } else {
            // So far so good, let's proceed
            if (twi_op == TWI_WRITE) {
                // If we're writing a single byte, let's just send it
                twi_state = TWI_SENDING_DATA;
                TWDR = twi_eeprom_data;
                TWCR = _BV(TWINT) | _BV(TWEN) | _BV(TWIE);
            } else 
            if (twi_op == TWI_PAGE_WRITE) {
                // If we're writing a page, let's send the data from
                // the queue
                twi_state = TWI_SENDING_DATA;
                TWDR = event_queue.twi_page_buf[twi_eeprom_data_index];
                twi_eeprom_inc_index();
                TWCR = _BV(TWINT) | _BV(TWEN) | _BV(TWIE);
            } else {
                // If we're reading, let's do a restart
                twi_state = TWI_SENDING_RESTART;
                TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN) | _BV(TWIE);
            }
        }
    } else if (twi_state==TWI_SENDING_DATA) {
        if (TW_STATUS == TW_MT_DATA_ACK) {
            unsigned char done=1;
            if (twi_op == TWI_PAGE_WRITE) {
                if (twi_eeprom_data_len) {
                    TWDR = event_queue.twi_page_buf[twi_eeprom_data_index];
                    twi_eeprom_inc_index();
                    TWCR = _BV(TWINT) | _BV(TWEN) | _BV(TWIE);
                    done=0;
                }
            } 
            if (done) {
                //twi_state=TWI_IDLE;
                twi_guard=10;
                twi_fail_state=TWI_SUCCESS;
                // Send STOP condition to finish transmission
                TWCR = _BV(TWINT) | _BV(TWEN) | _BV(TWSTO);
            }
        } else {
            twi_fail();
        }
    } else if (twi_state==TWI_SENDING_RESTART) {
        if (TW_STATUS == TW_MT_ARB_LOST) {
            twi_retry();
        } else if ( (TW_STATUS != TW_REP_START) && (TW_STATUS != TW_START)) {
            twi_fail();
        } else {
            twi_state=TWI_SENDING_SLA_R;
            TWDR = 0xA1;
            TWCR = _BV(TWINT) | _BV(TWEN) | _BV(TWIE);
        }
    } else if (twi_state==TWI_SENDING_SLA_R) {
        if (TW_STATUS == TW_MR_SLA_NACK) {
            twi_retry();
        } else if (TW_STATUS == TW_MR_ARB_LOST) {
            twi_state=TWI_SENDING_START;
            TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN) | _BV(TWIE);
        } else if (TW_STATUS != TW_MR_SLA_ACK) {
            // Error, send stop condition
            twi_fail();
        } else {
            // So far so good, let's receive the data
            twi_state = TWI_READING_DATA;
            TWCR = _BV(TWINT) | _BV(TWEN) | _BV(TWIE) | _BV(TWEA);
        }
    } else if (twi_state==TWI_READING_DATA) {
        if (TW_STATUS == TW_MR_DATA_ACK) {
            *twi_eeprom_data_ptr++ = TWDR;
            if (--twi_eeprom_data_len) {
                TWCR = _BV(TWINT) | _BV(TWEN) | _BV(TWIE) | _BV(TWEA);
            } else {
                twi_guard=2;
                twi_fail_state=TWI_SUCCESS;
                TWCR = _BV(TWINT) | _BV(TWEN) | _BV(TWSTO);
            }
        } else {
            twi_fail();
        }
    }
}

void twi_retry(void)
{
    if (--twi_retry_count) {
        twi_state = TWI_SENDING_START;
        TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN) | _BV(TWIE);
    } else {
        twi_fail();
    }
}

void twi_fail(void)
{
    twi_fail_state=twi_state;
    //twi_state=TWI_IDLE;
    twi_guard=2;
    // Send stop condition
    TWCR = _BV(TWINT) | _BV(TWEN) | _BV(TWSTO);
}

unsigned char twi_is_busy(void)
{
    return twi_state!=TWI_IDLE;
}

#ifdef HAVE_XTRA_CMDS

void xmit_dump(void)
{
   if (twi_is_busy() || (fifo_head != fifo_tail)) return;
   if (hexdump_state==HEXDUMP_XMIT_SPACE) {
       if (hexdump_addr<hexdump_end) {
           do_xmit(twi_fail_state!=TWI_SUCCESS?'.':' ');
           twi_eeprom_read_byte(hexdump_addr);
           hexdump_state=HEXDUMP_XMIT_HI_DATA_NIBBLE;
        } else {
           hexdump_state=HEXDUMP_IDLE;
        }
   } else if (hexdump_state==HEXDUMP_XMIT_HI_DATA_NIBBLE) {
       xmit_hex_nibble(twi_eeprom_data>>4);
       hexdump_state=HEXDUMP_XMIT_LO_DATA_NIBBLE;
   } else if (hexdump_state==HEXDUMP_XMIT_LO_DATA_NIBBLE) {
       xmit_hex_nibble(twi_eeprom_data);
       hexdump_addr++;
       if (!(hexdump_addr&15)) {
           hexdump_state=HEXDUMP_XMIT_CR;
       } else {
           hexdump_state=HEXDUMP_XMIT_SPACE;
       }
   } else if (hexdump_state==HEXDUMP_XMIT_CR) {
       do_xmit(13);
       hexdump_state=HEXDUMP_XMIT_LF;
   } else if (hexdump_state==HEXDUMP_XMIT_LF) {
       do_xmit(10);
       hexdump_state=HEXDUMP_XMIT_SPACE;
   }
}

void xmit_hex_nibble(unsigned char nibble)
{
    nibble&=15;
    nibble+='0';
    if (nibble>'9') nibble+=7;
    do_xmit(nibble);
}

#endif // HAVE_XTRA_CMDS

#endif // HAVE_TWI

/* -------------------------------- LCD Support --------------------------- */

#ifdef HAVE_LCD

/* you shouldn't need to change anything below this line */

typedef unsigned char  u08;
typedef unsigned short u16;

/* instruction register bit positions */
#define LCD_CLR             0      /* DB0: clear display */
#define LCD_HOME            1      /* DB1: return to home position */
#define LCD_ENTRY_MODE      2      /* DB2: set entry mode */
#define LCD_ENTRY_INC       1      /*   DB1: 1=increment, 0=decrement  */
#define LCD_ENTRY_SHIFT     0      /*   DB2: 1=display shift on        */
#define LCD_ON              3      /* DB3: turn lcd/cursor on */
#define LCD_ON_DISPLAY      2      /*   DB2: turn display on */
#define LCD_ON_CURSOR       1      /*   DB1: turn cursor on */
#define LCD_ON_BLINK        0      /*     DB0: blinking cursor ? */
#define LCD_MOVE            4      /* DB4: move cursor/display */
#define LCD_MOVE_DISP       3      /*   DB3: move display (0-> cursor) ? */
#define LCD_MOVE_RIGHT      2      /*   DB2: move right (0-> left) ? */
#define LCD_FUNCTION        5      /* DB5: function set */
#define LCD_FUNCTION_8BIT   4      /*   DB4: set 8BIT mode (0->4BIT mode) */
#define LCD_FUNCTION_2LINES 3      /*   DB3: two lines (0->one line) */
#define LCD_FUNCTION_10DOTS 2      /*   DB2: 5x10 font (0->5x7 font) */
#define LCD_CGRAM           6      /* DB6: set CG RAM address */
#define LCD_DDRAM           7      /* DB7: set DD RAM address */
#define LCD_BUSY            7      /* DB7: LCD is busy */

/* set entry mode: display shift on/off, dec/inc cursor move direction */
#define LCD_ENTRY_DEC            0x04   /* display shift off, dec cursor move dir */
#define LCD_ENTRY_DEC_SHIFT      0x05   /* display shift on,  dec cursor move dir */
#define LCD_ENTRY_INC_           0x06   /* display shift off, inc cursor move dir */
#define LCD_ENTRY_INC_SHIFT      0x07   /* display shift on,  inc cursor move dir */

/* move cursor/shift display */
#define LCD_MOVE_CURSOR_LEFT     0x10   /* move cursor left  (decrement)          */
#define LCD_MOVE_CURSOR_RIGHT    0x14   /* move cursor right (increment)          */
#define LCD_MOVE_DISP_LEFT       0x18   /* shift display left                     */
#define LCD_MOVE_DISP_RIGHT      0x1C   /* shift display right                    */

/* function set: set interface data length and number of display lines */
#define LCD_FUNCTION_4BIT_1LINE  0x20   /* 4-bit interface, single line, 5x7 dots */
#define LCD_FUNCTION_4BIT_2LINES 0x28   /* 4-bit interface, dual line,   5x7 dots */
#define LCD_FUNCTION_8BIT_1LINE  0x30   /* 8-bit interface, single line, 5x7 dots */
#define LCD_FUNCTION_8BIT_2LINES 0x38   /* 8-bit interface, dual line,   5x7 dots */


#define LCD_MODE_DEFAULT     ((1<<LCD_ENTRY_MODE) | (1<<LCD_ENTRY_INC) )

#define LCD_RS_DDR       DDRD   /* ddr for RS line */
#define LCD_RS_PORT      PORTD  /* port for RS line */
#define LCD_RS_PIN       PD3
#define LCD_E_DDR        DDRB  /* ddr for Enable line */
#define LCD_E_PORT       PORTB  /* port for Enable line */
#define LCD_E_PIN        PB2

/* normally you do not change the following */
#define LCD_LINE_LENGTH  0x40     /* internal line length */
#define LCD_START_LINE1  0x00     /* DDRAM address of first char of line 1 */
#define LCD_START_LINE2  0x40     /* DDRAM address of first char of line 2 */
#define LCD_START_LINE3  0x14     /* DDRAM address of first char of line 3 */
#define LCD_START_LINE4  0x54     /* DDRAM address of first char of line 4 */

#if LCD_LINES==1
#define LCD_FUNCTION_DEFAULT    LCD_FUNCTION_4BIT_1LINE
#else
#define LCD_FUNCTION_DEFAULT    LCD_FUNCTION_4BIT_2LINES
#endif

#define lcd_e_high()     LCD_E_PORT |=  _BV(LCD_E_PIN)
#define lcd_e_low()      LCD_E_PORT &= ~_BV(LCD_E_PIN) 
#define lcd_e_toggle()   toggle_e()

#define lcd_cmd_mode()   LCD_RS_PORT &= ~_BV(LCD_RS_PIN) // RS=0  command mode
#define lcd_data_mode()  LCD_RS_PORT |=  _BV(LCD_RS_PIN) // RS=1  data mode

static void inline lcd_data_port_in(void)
{
    DDRD &= 0x8B;
}

static void inline lcd_data_port_out(void)
{
    DDRD |= 0x74;
}

static void inline lcd_out_high(u08 d)
{
    PORTD = (PORTD & 0x8B) | (d & 0x70) | ((d&0x80)?4:0);
}

static void inline lcd_out_low(u08 d)
{
    lcd_out_high(d<<4);
}

static void toggle_e(void)
/* toggle Enable Pin */
{
    lcd_e_high();
    asm("nop"); asm("nop"); asm("nop"); asm("nop");
//    asm("nop"); asm("nop"); asm("nop"); asm("nop");
    lcd_e_low();
}

void lcd_write(unsigned char data, unsigned char rs)
{
    /* configure data pins as output */
    lcd_data_port_out();

    /* output high nibble first */

    lcd_out_high(data);

    if (rs)
        lcd_data_mode();    /* RS=1: write data            */
    else
        lcd_cmd_mode(); /* RS=0: write instruction     */
    lcd_e_toggle();

    /* output low nibble */
    lcd_out_low(data);

    lcd_e_toggle();

    /* all data pins high (inactive) */
    lcd_data_port_in();
}

/*
** PUBLIC FUNCTIONS 
*/

void lcd_command(u08 cmd)
/* send commando <cmd> to LCD */
{
    //lcd_waitbusy();
    lcd_write(cmd, 0);
}

void lcd_clrscr(void)
/* clear lcd and set cursor to home position */
{
    lcd_command(_BV(LCD_CLR));
}

void lcd_home(void)
/* set cursor to home position */
{
    lcd_command(_BV(LCD_HOME));
}

void lcd_init(unsigned char dispAttr)
/* initialize display and select type of cursor */
/* dispAttr: LCD_DISP_OFF, LCD_DISP_ON, LCD_DISP_ON_CURSOR, LCD_DISP_CURSOR_BLINK */
{
    /*------ Initialize lcd to 4 bit i/o mode -------*/

    lcd_data_port_out();    /* all data port bits as output */
    LCD_RS_DDR |= _BV(LCD_RS_PIN);
    LCD_E_DDR  |= _BV(LCD_E_PIN);

    _delay_ms(16);    /* wait 16ms or more after power-on       */

    /* initial write to lcd is 8bit */
    lcd_out_high(LCD_FUNCTION_8BIT_1LINE);
    lcd_e_toggle();
    _delay_us(4992);     /* delay, busy flag can't be checked here */

    lcd_out_high(LCD_FUNCTION_8BIT_1LINE);
    lcd_e_toggle();
    _delay_us(64);       /* delay, busy flag can't be checked here */

    lcd_out_high(LCD_FUNCTION_8BIT_1LINE);
    lcd_e_toggle();
    _delay_us(64);       /* delay, busy flag can't be checked here */

    lcd_out_high(LCD_FUNCTION_4BIT_1LINE);  /* set IO mode to 4bit */
    lcd_e_toggle();

    /* from now the lcd only accepts 4 bit I/O, we can use lcd_command() */
    lcd_command(LCD_FUNCTION_DEFAULT);  /* function set: display lines  */
    _delay_ms(1);
    lcd_command(LCD_DISP_OFF);  /* display off                  */
    _delay_ms(1);
    lcd_clrscr();       /* display clear                */
    _delay_ms(1);
    lcd_command(LCD_MODE_DEFAULT);  /* set entry mode               */
    _delay_ms(1);
    lcd_command(dispAttr);  /* display/cursor control       */
    _delay_ms(1);
}

#endif // HAVE_LCD

void execute_command(void)
{
    cmd_in_execution=2;
    unsigned char vect=0;
    prog_char *p=cmdnames;
    prog_char *line=cmdline;
    cmdline[cmdline_cursor]=0;
    // Ignore leading spaces
    while (*line==' ') line++,cmdline_cursor--;
    for (;cmdline_cursor;) {
        if (!pgm_read_byte(p)) {
            do_xmit_pstr("Unrecognized command");
            break;
        }
        if (!strncmp_P(line,p,strlen_P(p))) {
            unsigned char argc=0;
            char *argv[MAXARGS];
            while (*line && *line!=' ') line++;
            for (;*line;line++) {
                if (*line==' ') {
                    *line=0;
                } else {
                    if (!line[-1] && argc<MAXARGS) argv[argc++]=line;
                }
            }
            // Jumps to the function that handles that command, passing
            // the number of arguments and the separated words in the
            // command line. This code is a bit unreadable, but it simply
            // fetches the function address from the cmdvects array (indexed
            // by vect), converts the word to a function pointer and calls
            // the address pointed to by this pointer with argc and argv as
            // addresses.
            ((void (*)(int,char **))pgm_read_word(cmdvects+vect))(argc,argv);
            break;
        }
        while (pgm_read_byte(p)) p++;
        p++; vect++;
    }
    do_xmit_pstr("\r\n> ");
    cmd_in_execution=0; cmdline_cursor=0;
}

void cmd_date(unsigned char argc, char **argv)
{
    if (argc!=3) {
        usage("date yy mm dd");
        return;
    }
    set_time_date(3,argv,YEAR);
}

void cmd_time(unsigned char argc, char **argv)
{
    if (argc!=3 && argc!=2) {
        usage("time hh mm [ss]");
        return;
    }
    set_time_date(argc,argv,HOUR);
    grace_time=2;
}

void set_time_date(unsigned char argc, char **argv, unsigned char base)
{
    unsigned char i;
    flags |= FLAG_CLOCK_ADJUST;
    for (i=0;i<argc;i++) {
        rtc_crystal[base-i]=atoi(argv[i]);
    }
    if (base==HOUR) {
        rtc_crystal[SUBSEC]=0;
        if (argc==2) rtc_crystal[SEC]=0;
    }
    memcpy((char *)rtc_mains,(char *)rtc_crystal,sizeof(rtc_crystal));
    flags &= ~FLAG_CLOCK_ADJUST;
}

void cmd_log(unsigned char argc, char **argv)
{
    flags |= FLAG_SERIAL_LOG_ENABLED;
}

void cmd_clear(unsigned char argc, char **argv)
{
    cli();
    slow_semicycles=0;
    fast_semicycles=0;
    sag_semicycles=0;
    surge_semicycles=0;
    nevents=-1;
    within_event=FALSE;
    unflushed_events=0;
    first_unflushed=0;
    event_tail=0;
    lcd_current=0;
    grace_time=240;
    beep_cycles=0;
    semicycles=0;
    BUZZER_PORT &= ~_BV(BUZZER_PIN);
    sei();
    memset(lcd,' ',sizeof(lcd));
}

void cmd_event(unsigned char argc, char **argv)
{
    if (argc!=1) {
       enable_serial_override();
       xmit_int_dp(nevents+1,5,-1);
       xmit_pstr(" events recorded\r\n");
       disable_serial_override();
       return;
    }
    unsigned short e=atoi(argv[0]);
    struct event_st *evt=get_event(e);
    if (!evt) {
        do_xmit_pstr("No such event\r\n"); return;
    }
    enable_serial_override();
    xmit('['); xmit_int_dp(evt->type,3,-1); xmit_pstr("] ");
    xmit_clock(evt->timestamp,7,0); xmit(' ');
    xmit_elapsed(evt->duration);
    xmit_pstr("\r\n");
    disable_serial_override();
}

void cmd_reboot(unsigned char argc, char **argv)
{
    cli();
    WDTCR |= _BV(WDE);
    for (;;);
}

void cmd_param(unsigned char argc, char **argv)
{
    if (argc!=1 && argc!=2) {
        usage("param index [value]"); return;
    }
    unsigned char i=atoi(argv[0]);
    if (argc==2) {
        params[i]=atoi(argv[1]);
    }
    enable_serial_override();
    xmit_int_dp(i,2,-1); xmit('=');
    xmit_int_dp(params[i],5,-1);
    xmit_pstr("\r\n");
    disable_serial_override();
}

void cmd_ver(unsigned char argc, char **argv)
{
    do_xmit_pstr("Power Minder v0.4 - (C) 2008 by Kiko built on "
     __DATE__ " " __TIME__ "\r\n");
}

void show_usage(prog_char *p)
{
    do_xmit_pstr("Usage: ");
    do_xmit_str_p(p);
    do_xmit_pstr("\r\n");
}

/* ----------------- Extra Commands for Debugging Purposes ---------------- */

#ifdef HAVE_XTRA_CMDS

void cmd_poke(unsigned char argc, char **argv)
{
    if (argc!=2) {
        usage("poke addr data"); return;
    }
    unsigned short addr=atoi(argv[0]);
    unsigned char data=atoi(argv[1]);
    twi_eeprom_write_byte(addr,data);
}

#ifdef HAVE_TWI

void cmd_dump(unsigned char argc, char **argv)
{
    if (argc!=1 && argc!=2) {
        usage("dump addr [size]"); return;
    }
    unsigned short size=16;
    hexdump_addr=atoi(argv[0]);
    if (argc==2) size=atoi(argv[1]);
    hexdump_end=hexdump_addr+size;
    hexdump_state=HEXDUMP_XMIT_CR;
}

#endif // HAVE_TWI

void cmd_fill(unsigned char argc, char **argv)
{
    if (argc!=3) {
        usage("fill addr data inc"); return;
    }
    unsigned short addr=atoi(argv[0]);
    unsigned char data=atoi(argv[1]);
    unsigned char inc=atoi(argv[2]);
    unsigned char n;

    for (n=0;n<TWI_PAGE_SIZE;n++) {
         event_queue.twi_page_buf[n]=data;
         data+=inc;
    }
    twi_eeprom_write_page(addr,0);
}

void cmd_start(unsigned char argc, char **argv)
{
    if (argc!=1) {
        usage("start event-type"); return;
    }
    start_new_event(atoi(argv[0]));
}

void cmd_finish(unsigned char argc, char **argv)
{
    finish_event();
    enable_serial_override();
    xmit_int_dp(unflushed_events,2,-1);
    xmit('/');
    xmit_int_dp(event_tail,2,-1);
    xmit('/');
    xmit_int_dp(first_unflushed,2,-1);
    xmit_pstr("\r\n");
    disable_serial_override();
}

#endif // HAVE_XTRA_CMDS
